import numpy as np
import multiprocessing as mp
import warnings
from .utils import (getCounts, randomPartition,
                    copyProxyDictionary, computeSamplePvalue)


class PathwayRepresentation:
    """
    Methods to compute pathway and system representation in gene clusters.
    A system is a collection of pathways and each pathway a collection of genes.
    """
    def __init__(self, clusters: dict, gene_pathways: dict,
                 system_pathways: dict = None) -> None:
        self._gene_pathways = gene_pathways
        self._pathway_sizes = getCounts(
            [p for pathways in self._gene_pathways.values() for p in pathways]
        )
        self._pathway_ids = list(self._pathway_sizes.keys())
        self._system_pathways = system_pathways
        if system_pathways is not None:
            self.checkSystemPathways()
        self._clusters = clusters
        self._cluster_ids = list(self._clusters.keys())

    def getPathwaysFromGeneList(self, gene_list: list) -> list:
        """
        Get list of pathways (not unique) represented in gene list
        """
        genes_with_assigned_pathway = [g for g in gene_list if g in self._gene_pathways]
        return [p for g in genes_with_assigned_pathway for p in self._gene_pathways[g]]

    def initializeClusterRepresentationDict(self) -> dict:
        """
        Initialize cluster pathway representation dict with keys equal to pathways
        and  values equal to 0.
        """
        return {pathway: 0 for pathway in self._pathway_ids}

    def initializeClustersRepresentationDict(self) -> dict:
        """
        Initialize a cluster representation dictionary with 0s. Dict of dicts
        with keys equal to cluster ids and pathway names within.
        """
        return {
                cluster_id: self.initializeClusterRepresentationDict()
                for cluster_id in self._cluster_ids
        }

    def computeClusterPathwayRepresentation(self, cluster_pathway_sizes: dict) -> dict:
        pathway_representation = self.initializeClusterRepresentationDict()
        for pathway, total_counts in self._pathway_sizes.items():
            if pathway in cluster_pathway_sizes.keys():
                pathway_representation[pathway] = cluster_pathway_sizes[pathway] / total_counts
        return pathway_representation

    def getClustersPathwayRepresentation(self, clusters: dict) -> dict:
        """
        Compute fraction of genes of each pathway represented in each cluster
        gene_pathways: dict with keys being gene ids and values are lists of pathways (str)
        to which each gene belongs.
        """
        clusters_pathway_representation = {}
        for cluster_id, cluster in clusters.items():
            cluster_pathway_sizes = getCounts(
                self.getPathwaysFromGeneList(cluster)
            )
            clusters_pathway_representation[cluster_id] = self.computeClusterPathwayRepresentation(
                cluster_pathway_sizes)
        return clusters_pathway_representation

    def checkSystemPathways(self):
        """
        Check if all pathways defined within systems are included in the
        pathway set gathered for current gene set. Used to check if input
        systems_pathway_dict is consistent with pathway data.
        """
        if self._system_pathways is None:
            raise ValueError('Undefined system_pathways dictionary')
        system_included_pathways = {}
        ill_defined_systems = []
        for system, pathways in self._system_pathways.items():
            included_pathways = [
                pathway for pathway in pathways if pathway in self._pathway_ids
            ]
            system_included_pathways[system] = included_pathways
            if len(included_pathways) < len(pathways):
                ill_defined_systems.append(system)
        if ill_defined_systems:
            warnings.warn('Some systems contained pathways without assigned genes. Removing these pathways from dictionary')
        self._system_pathways = system_included_pathways

    def getClustersSystemRepresentation(self,
                                        clusters_pathway_representation: dict) -> dict:
        """
        Compute system representation out of pathway
        representation values (a system is a set of pathways or subsystems)

        R_sys = (1/N_sys) * Sum_path (R_path * N_path), for all pathways in system,
        where: N_sys, number of genes in system
               R_sys, system representation in cluster (genes in cluster/total genes in system)
               N_path, number of genes in pathway
               R_path, pathway representation in cluster (genes in cluster/total genes in pathway)
        """
        if self._system_pathways is None:
            raise ValueError('Undefined system_pathways dictionary')
        clusters_system_representation = {}
        for cluster_id, cluster_rep in clusters_pathway_representation.items():
            clusters_system_representation[cluster_id] = {}
            for system_id, pathways in self._system_pathways.items():
                system_size = sum([self._pathway_sizes[pathway] for pathway in pathways])
                system_rep = (1 / system_size) * sum([cluster_rep[pathway] * self._pathway_sizes[pathway] for pathway in pathways])
                clusters_system_representation[cluster_id][system_id] = system_rep
        return clusters_system_representation


class ClusterPermutator(PathwayRepresentation):
    """
    Methods to run a permutation analysis on gene clusters
    pathway_representation: class instance of PathwayRepresentation
    NOTE: The inits of these two classes are redundant. Improve it.
    """
    def __init__(self, clusters: dict, gene_pathways: dict,
                 system_pathways: dict = None) -> None:
        super().__init__(clusters, gene_pathways, system_pathways)
        self._cluster_pathway_representation = self.getClustersPathwayRepresentation(
            self._clusters)
        if system_pathways is not None:
            self._cluster_system_representation = self.getClustersSystemRepresentation(
                self._cluster_pathway_representation
            )
        self._gene_list = [g for c in self._clusters.values() for g in c]
        self._cluster_sizes = [len(v) for v in self._clusters.values()]
        self._cluster_representation_p_values = self.initializeClustersRepresentationDict()
        self._sample_size = None

    def getPermutedClusterPathwayRepresentation(self, i=None):
        """
        Compute pathway representation in permuted clusters
        (Ignore 'i' argument, required by Multiprocessing.map)
        """
        permuted_clusters = dict(
            zip(self._cluster_ids, randomPartition(self._gene_list, self._cluster_sizes))
        )
        return self.getClustersPathwayRepresentation(permuted_clusters)

    def _aggregatePermutationPathwayResults(self, samples: list) -> dict:
        aggregated_samples = {
            cluster_id: {pathway: []
                         for pathway in self._pathway_ids}
            for cluster_id in self._cluster_ids
        }
        for sample in samples:
            for cluster_id in self._cluster_ids:
                for pathway in self._pathway_ids:
                    aggregated_samples[cluster_id][pathway].append(sample[cluster_id][pathway])
        return aggregated_samples

    def _aggregatePermutationSystemResults(self, samples: list) -> dict:
        aggregated_samples = {
            cluster_id: {system: []
                         for system in self._system_pathways.keys()}
            for cluster_id in self._cluster_ids
        }
        for sample in samples:
            for cluster_id in self._cluster_ids:
                for system in self._system_pathways.keys():
                    aggregated_samples[cluster_id][system].append(sample[cluster_id][system])
        return aggregated_samples

    def aggregatePermutationSampleResults(self, samples: list, system_type='pathway') -> dict:
        """
        Aggregate permutation sample results (generated via multiprocessing)
        into a single object
        """
        if 'pathway' in system_type.lower():
            return self._aggregatePermutationPathwayResults(samples)
        else:
            return self._aggregatePermutationSystemResults(samples)

    def _sortPathwaysByPvalue(self, cluster_pathway_pvalue: dict) -> dict:
        """
        Sort pathways within a cluster based on their sample pvalue.
        cluster_pathway_pvalue, dictionary where keys are pathway names
        and values are tuples: (representation, p-value)
        """
        pathways = cluster_pathway_pvalue
        sorted_keys = np.array(
                               list(pathways.keys())
        )[np.argsort([pvalue for rep, pvalue in pathways.values()])]
        return {k: pathways[k] for k in sorted_keys}

    def computePermutationSamplePvalue(self, aggregated_samples: dict,
                                       pathway_representation: dict) -> dict:
        """
        Compute sample p-value for each pathway and cluster
        """
        rep_pvalues = {cluster_id: {} for cluster_id in pathway_representation.keys()}
        for cluster_id, rep_values in pathway_representation.items():
            pathways_rep = {}
            for pathway_id, rep_value in rep_values.items():
                rep_distribution = aggregated_samples[cluster_id][pathway_id]
                p_value = computeSamplePvalue(rep_distribution, rep_value)
                pathways_rep[pathway_id] = (rep_value, p_value)
            rep_pvalues[cluster_id] = self._sortPathwaysByPvalue(pathways_rep)
        return rep_pvalues

    def sampleClusterPermutationSpace(self, sample_size=1000, n_processes=None) -> list:
        """
        Permute genes in clusters n times.
        NOTE: Postprocessing time can be optimized, e.g., by choosing list of lists
        instead of dictionary to store pathway representation sample. Currently it takes
        over a minute to process samples (when both pathway and system representation
        are computed)
        """
        self._sample_size = sample_size
        if n_processes is None:
            n_processes = mp.cpu_count() - 1
        pool = mp.Pool(n_processes)
        samples = pool.map(
            self.getPermutedClusterPathwayRepresentation,
            [() for _ in range(sample_size)]
        )
        pool.close()
        print('Finished permutation sampling')
        pathway_rep_result = copyProxyDictionary(
            self.computePermutationSamplePvalue(self.aggregatePermutationSampleResults(
                samples, system_type='pathway'), self._cluster_pathway_representation)
            )

        if self._system_pathways is None:
            return pathway_rep_result
        else:
            system_samples = [
                self.getClustersSystemRepresentation(sample) for sample in samples
                ]
            system_rep_result = copyProxyDictionary(self.computePermutationSamplePvalue(
                self.aggregatePermutationSampleResults(system_samples, system_type='system'),
                self._cluster_system_representation)
                )
            return {'pathway': pathway_rep_result, 'system': system_rep_result}

    # Shared memory in multiprocessing (not very efficient)

    # def getPermutedClusterPathwayRepresentationPvalue(self, shared_res_dict):
    #     """
    #     Compute pathway representation in permuted clusters and compare with
    #     actual value
    #     """
    #     permuted_representation = self.getPermutedClusterPathwayRepresentation()
    #     for cluster_id, cluster in self._cluster_pathway_representation.items():
    #         for pathway, representation in cluster.items():
    #             if permuted_representation[cluster_id][pathway] >= representation:
    #                 self._cluster_representation_p_values[cluster_id][pathway] += 1 / self._sample_size
    #
    # def sampleClusterPermutationSpaceSharedMemory(self,
    #                                               sample_size=1000,
    #                                               n_processes=None) -> list:
    #     """
    #     Permute genes in clusters n times
    #     NOTE: Shared memory doesn't seem to work properly. When using a shared dict
    #     one process heavely uses CPU and the others do not. Executing time is far greater
    #     than multiprocessing with no shared memory.
    #     """
    #     self._sample_size = sample_size
    #     manager = mp.Manager()
    #     shared_res_dict = {
    #                        cluster_id: manager.dict(self.initializeClusterRepresentationDict())
    #                        for cluster_id in self._cluster_ids
    #                        }
    #     self._cluster_representation_p_values = shared_res_dict
    #
    #     if n_processes is None:
    #         n_processes = mp.cpu_count() - 1
    #
    #     pool = mp.Pool(n_processes)
    #     pool.map(
    #         self.getPermutedClusterPathwayRepresentationPvalue,
    #         [_ for _ in range(sample_size)]
    #     )
    #     pool.close()
    #     return shared_res_dict



from matplotlib import pyplot as plt
import pandas as pd
import numpy as np
import re
from staticinteract import StaticInteract, DropDownWidget, RangeWidget


class PathwayPlotter:
    """
    Methods  to plot pathway representation results
    """
    def __init(self):
        pass


def extractKoPathwayName(Ko_str):
    return re.sub('\[.*?\]', '', Ko_str).strip()


# Plotting functions
def plotClusterData(pdata, cluster, ax=None, cluster_id=None):
    pdata[pdata.index.isin(cluster)].transpose().plot(
        legend=False, title=f'{cluster_id}, size={len(cluster)}',
        ax=ax, color='#9a9a9a', linewidth=0.8,
        marker='.', markerfacecolor='#ee9929', markersize=12)


def plotKEGGFrequencies(data: dict, color=None, axis=None):
    """
    Bar plot of sorted KEGG systems or subsystems
    data: dictionary wth keys being system or subsystem names and
    values, their frequencies/representation.
    """
    if color is None:
        color = 'C0'
    pvalues = [v[1] for v in data.values()]
    clean_name_data = 100 * pd.Series(
        {extractKoPathwayName(k): data[k][0] for k in data.keys()}
    )
    ax = clean_name_data.plot.bar(figsize=(12, 8), color=color, ax=axis)
    for i, p in enumerate(ax.patches):
        ax.annotate(f'({pvalues[i]:.4f})', (p.get_x() * 1.005, p.get_height() * 1.006))


def plotSystemsAndSubsystemsWebPage(clusters, cluster_data, permutation_results,
                                    plot_first_N=10, color=None,
                                    img_folder_name=None):
    """
    NOTE: Increase tick labels font size!
    """

    plt.rcParams.update({'figure.max_open_warning': 0})
    if color is None:
        color = 'C0'
    if img_folder_name is None:
        img_folder_name = 'iplot'

    system_types = ['system', 'pathway']
    cluster_ids = list(clusters.keys())

    def plot_fun(system_type, cluster_id):

        fig, ax = plt.subplot_mosaic(
            """
            A
            B
            """,
            gridspec_kw = {"height_ratios": [0.7, 1]}
        )

        ax['B'].set_ylabel('Pathway representation (%)')
        ax['B'].set_title(f'{system_type}s (sample p-value)')

        if cluster_id in clusters.keys():
            p_Data = permutation_results
            kdata = {k: v
                     for k,v in p_Data[system_type][cluster_id].items()}
            if len(kdata) > plot_first_N:
                kdata = {k: kdata[k] for k in list(kdata.keys())[:10]}
            plotClusterData(cluster_data, clusters[cluster_id],
                            ax['A'], cluster_id)
            plotKEGGFrequencies(kdata,
                                color=color, axis=ax['B'])
        fig.set_figwidth(20)
        fig.set_figheight(20)
        return fig

    i_fig = StaticInteract(plot_fun,
                           system_type=DropDownWidget(system_types,
                                        description='Level'),
                           cluster_id=DropDownWidget(cluster_ids,
                                        description='Cluster ID'),
                           interact_name=img_folder_name)
    return i_fig


def getRepresentationStackedPlotData(p_paths, pvalue_cutoff=1):
    """
    Make pandas dataframe of pathway representation across
    clusters
    """
    unique_systems = np.unique([
        system for c in p_paths['system'].values()
        for system in c.keys()
    ])
    unique_subsystems = np.unique([
        subsystem
        for c in p_paths['pathway'].values()
        for subsystem in c.keys()
    ])

    rep_systems, rep_subsystems = {}, {}
    for system in unique_systems:
        rep_systems[system] = []
        for cluster_id, values in p_paths['system'].items():
            rep, pvalue = values[system]
            if pvalue < pvalue_cutoff:
                rep_systems[system].append(rep)
            else:
                rep_systems[system].append(0.0)
        # Add rep outside clusters
        rep_systems[system].append(1 - sum(rep_systems[system]))

    for subsystem in unique_subsystems:
        rep_subsystems[subsystem] = []
        for cluster_id, values in p_paths['pathway'].items():
            rep, pvalue = values[subsystem]
            if pvalue < pvalue_cutoff:
                rep_subsystems[subsystem].append(rep)
            else:
                rep_subsystems[subsystem].append(0.0)
        # Add rep outside clusters
        rep_subsystems[subsystem].append(1 - sum(rep_subsystems[subsystem]))

        rep_subsystems = {extractKoPathwayName(k): v
                          for k,v in rep_subsystems.items()}

    return {'system': rep_systems, 'pathway': rep_subsystems}


def plotSystemsAndSubsystemsStacked(permutation_results, cluster_colors, img_folder_name):
    """
    """
    plt.rcParams.update({'figure.max_open_warning': 0})
    if img_folder_name is None:
        img_folder_name = 'iplot'

    # Remove No cluster assigned data
    p_Data = {system_type: {system: value for system, value in data.items() if system != 'No_cluster_assigned'}
              for system_type, data in permutation_results.items()}

    system_types = ['system', 'pathway']

    def plot_fun(system_type, tpvalue_cutoff):
        fig = plt.figure(figsize=(10, 8))
        ax = plt.gca()
        rep_data = getRepresentationStackedPlotData(p_Data, tpvalue_cutoff)

        df = pd.DataFrame(rep_data[system_type],
                          index=list(p_Data['system'].keys()) + ['No cluster'])
        df = df.loc[:, df.loc['No cluster'] < 1.0].sort_values('No cluster', axis=1)
        ax = (100 * df).transpose().plot.bar(stacked=True, legend=True,
                                             figsize=(10,8),
                                             color=list(cluster_colors.values()),
                                             ax=ax)
        return fig

    i_fig = StaticInteract(plot_fun,
                           system_type=DropDownWidget(system_types,
                                        description='Level'),
                           tpvalue_cutoff=RangeWidget(min=0.05, max=1,
                                                     step=0.05,
                                                     default=0.1,
                                                     description='p-value cutoff'),
                           interact_name=img_folder_name)
    return i_fig



import re
import io
import numpy as np
import pandas as pd
import json
import requests
import warnings


def downloadKEGGpathwaysForID(KEGG_entry_ID: str) -> dict:
    """
    KEGG_estry_ID: genome entry ID or organism code within KEGG's database.
    """
    url = f'http://rest.kegg.jp/list/pathway/{KEGG_entry_ID}'
    text = requests.get(url).text
    columns = ['dok', 'pathway name']
    df = pd.read_csv(io.StringIO(text), sep='\t', header=None, names=columns)
    df = df.applymap(lambda s: s.replace('path:', ''))
    pathway_ids, pathway_names = df.values[:, 0], df.values[:, 1]
    return dict(zip(pathway_ids, pathway_names))


class KEGGPathwayParser:
    """
    Methods to parse KEGG pathway orthology
    """
    def __init__(self, KEGG_dict):
        """
        kid: KEGG identifier for desired organism
        """
        self._KEGG_dict = KEGG_dict
        self._gene_kos = None

    @classmethod
    def fromKEGGidentifier(cls, kid, only_curated_pathways=True):
        """
        Download KEGG functional orthology for specified KEGG identifier (kid).
        Pathways can be restricted to those that have been curated, as retrieved by the
        KEGG API.
        """
        kegg_pathways = cls._downloadKEGGorthology(kid)
        KEGG_dict = cls._getKEGGdict(kegg_pathways)
        if only_curated_pathways:
            return cls(cls.filterKEGGorthology(KEGG_dict, kid))
        else:
            return cls(KEGG_dict)

    @classmethod
    def fromJSON(self, path_to_json):
        """
        """

    @staticmethod
    def filterKEGGorthology(KEGG_dict, kid):
        """
        Filter organism-specific KEGG dict by curated pathways
        """
        filtered_KEGG_dict = {}
        curated_pathways = downloadKEGGpathwaysForID(kid)
        for supersystem_id, supersystem in KEGG_dict.items():
            filtered_KEGG_dict[supersystem_id] = {}
            for system_id, system in supersystem.items():
                pathways_to_keep = []
                for pathway_id, pathway in system.items():
                    if any([ko in pathway_id for ko in curated_pathways.keys()]):
                        pathways_to_keep.append(pathway_id)
                if pathways_to_keep:
                    filtered_KEGG_dict[supersystem_id][system_id] = {
                        pathway_id: pathway for pathway_id, pathway in system.items()
                        if pathway_id in pathways_to_keep
                    }
        return filtered_KEGG_dict

    @staticmethod
    def _downloadKEGGorthology(kid: str, path_to_save_json: str = None):
        """
        Download KEGG BRITE functional hierarchy
        Hierarchy can be downloaded for a specific organism, as long as it is
        listed in the KEGG database
        """
        if kid is None:
            kid = 'ko'
        url = f'http://rest.kegg.jp/get/br:{kid}00001/json'
        file = requests.get(url, allow_redirects=True)
        if file.status_code == 404:
            warnings.warn(f'Organism not found in KEGG. Downloading generic orthology')
            url = f'http://rest.kegg.jp/get/br:ko00001/json'
            file = requests.get(url, allow_redirects=True)
        kegg_pathways = json.loads(file.content)['children']
        if path_to_save_json is not None:
            with open(path_to_save_json, 'w') as outfile:
                json.dump(kegg_pathways, outfile)
        return kegg_pathways

    def _downloadKEGGgenePathwaysForID(self, KEGG_entry_ID: str) -> dict:
        """
        KEGG_entry_ID: genome entry ID or organism code within KEGG's database.
        """
        url = f'http://rest.kegg.jp/link/{KEGG_entry_ID}/pathway'
        text = requests.get(url).text
        columns = ['ko', 'gene_id']
        df = pd.read_csv(io.StringIO(text), sep='\t', header=None, names=columns)
        kid = re.match('^[^:]*', df.iloc[0, 1]).group()
        df = df.applymap(lambda s: re.sub('^[^:]*:', '', s))
        df['ko'] = df['ko'].map(lambda s: s.replace(kid, 'ko'))
        df = df.reindex(columns=['gene_id', 'ko'])
        genes, kos = df.values[:, 0], df.values[:, 1]
        return dict(zip(genes, kos))

    def getKEGGgenePathwaysOLD(self, KEGG_entry_ID: str) -> dict:
        """
        KEGG_estry_ID: genome entry ID or organism code within KEGG's database.
        """
        gene_pathways, gene_systems = {}, {}
        if self._gene_kos is None:
            self._gene_kos = self._downloadKEGGgenePathwaysForID(KEGG_entry_ID)

        for gene_id, ko in self._gene_kos.items():
            try:
                ko_name = self.getPathwayOrthologyFromKO(ko)
            except Exception:
                ko_name = {'subsystem': '', 'system': ''}
            gene_pathways[gene_id] = ko_name['subsystem']
            gene_systems[gene_id] = ko_name['system']
        return (gene_pathways, gene_systems)

    @staticmethod
    def extractKoID(KEGG_pathway_name: str) -> str:
        """
        Extract KEGG's ko identifier from pathway name string.
        """
        try:
            return 'ko' + re.search('ko(.*?)\]', KEGG_pathway_name).group(1)
        except Exception:
            return ''

    @staticmethod
    def simplifyKEGGpathwayName(KEGG_pathway_name: str) -> str:
        """
        Remove ko identifier from KEGG pathway name
        """
        return re.sub('\d{5}', '', re.sub('\[.*?\]', '', KEGG_pathway_name)).strip()

    def parseKEGGPathwayJSON(self, path_to_json: str, simplified_names=False,
                             supersystems_to_keep: list=None) -> dict:
        """
        Parse KEGG pathway JSON into a nested dictionary
        """
        with open(path_to_json) as json_file:
            kegg_pathways = json.load(json_file)['children']
        if supersystems_to_keep is not None:
            kegg_pathways = [supersystem for supersystem in kegg_pathways
                             if supersystem['name'] in supersystems_to_keep]

        KEGG_dict = self._getKEGGdict(kegg_pathways)
        ko_orthology_dict = self._getKoOrthologyDict(kegg_pathways)
        return (KEGG_dict, ko_orthology_dict)

    @staticmethod
    def _getKEGGdict(kegg_pathways: dict) -> dict:
        """
        Obtain a simplified python dictionry from the direct JSON load object.
        """
        KEGG_dict = {}
        for supersystem in kegg_pathways:
            KEGG_dict[supersystem['name']] = {}
            for system in supersystem['children']:
                KEGG_dict[supersystem['name']][system['name']] = {}
                for pathway in system['children']:
                    if 'children' in pathway.keys():
                        KEGG_dict[supersystem['name']][
                            system['name']][pathway['name']] = []
                        for gene in pathway['children']:
                            KEGG_dict[supersystem['name']][
                                system['name']][pathway['name']].append(gene['name'])
        return KEGG_dict

    def _getGeneListFromKEGGorthology(self) -> list:
        """
        Retrieve list of included genes in (organism-specific) KEGG orthology
        """
        genes = []
        for supersystem_id, supersystem in self._KEGG_dict.items():
            for system_id, system in supersystem.items():
                for pathway_id, pathway in system.items():
                    for gene_str in pathway:
                        gene_id = gene_str.split(' ')[0]
                        genes.append(gene_id)
        return np.unique(genes).tolist()

    def getGeneInfoFromKEGGorthology(self) -> list:
        """
        Retrieve gene product info of included genes in (organism-specific) KEGG orthology
        """
        gene_info = {}
        for supersystem_id, supersystem in self._KEGG_dict.items():
            for system_id, system in supersystem.items():
                for pathway_id, pathway in system.items():
                    for gene_str in pathway:
                        gene_id = gene_str.split(' ')[0]
                        if gene_id not in gene_info.keys():
                            gene_meta = gene_str.split('\t')[1]
                            gene_info[gene_id] = gene_meta
        return gene_info

    def getGenePathways(self):
        """
        Extract assigned KEGG pathways and systems for each gene in organism-specific
        KEGG function orthology.
        """
        genes = self._getGeneListFromKEGGorthology()
        gene_pathways = {gene_id: [] for gene_id in genes}
        gene_systems = {gene_id: [] for gene_id in genes}

        for supersystem_id, supersystem in self._KEGG_dict.items():
            for system_id, system in supersystem.items():
                for pathway_id, pathway in system.items():
                    for gene_str in pathway:
                        gene_id = gene_str.split(' ')[0]
                        gene_pathways[gene_id].append(pathway_id)
                        gene_systems[gene_id].append(system_id)
        return (gene_pathways, gene_systems)

    def getSystemPathways(self):
        """
        Retrieve pathways (subsystems) within each KEGG system
        """
        system_pathways = {}
        for supersystem_id, supersystem in self._KEGG_dict.items():
            for system_id, system in supersystem.items():
                pathways = [path_id for path_id in system.keys()]
                if pathways:
                    system_pathways[system_id] = pathways
        return system_pathways

    def _getKoOrthologyDict(self, kegg_pathways: dict) -> dict:
        """
        Obtain dictionary containing KEGG pathway classification for each
        KEGG pathway id (koXXXXX)
        simplified_names: self.simplifyKEGGpathwayName
        """
        ko_orthology = {}

        for supersystem in kegg_pathways:
            for system in supersystem['children']:
                for subsystem in system['children']:
                    try:
                        ko_id = self.extractKoID(subsystem['name'])
                        ko_path_name = subsystem['name']
                    except Exception:
                        pass
                    ko_orthology[ko_id] = {
                        'subsystem': ko_path_name,
                        'system': system['name'],
                        'supersystem': supersystem['name']
                    }
        return ko_orthology

    def getPathwayOrthologyFromKO(self, ko_identifier: str) -> dict:
        """
        Extract KEGG orthology for given ko identifier
        subsystem (pathway) < system < supersystem)
        """
        try:
            return self._ko_orthology[ko_identifier]
        except Exception:
            raise ValueError('KO not in local database')

    def getGenePathwaysDictionaryFromEGGNOG(self, path_to_eggnog_data: str) -> dict:
        """
        Make dictionary where keys are gene IDs and values lists of pathways.
        """



from sklearn.metrics import silhouette_samples
import pandas as pd


def computeGeneSilhouettes(clusters, data):
    """
    Compute gene silhouettes for each gene in clusters
    """
    genes_in_clusters, cluster_labels = [], []
    for cluster_id, cluster in clusters.items():
        genes_in_clusters.extend(cluster)
        cluster_labels.extend([cluster_id for _ in range(len(cluster))])

    X = data.loc[genes_in_clusters, :].values
    sil_values = silhouette_samples(X, cluster_labels)

    return {gene_id: sil_values[i] for i, gene_id in enumerate(genes_in_clusters)}


def rankGenesWithinClusters(clusters, data):
    """
    Rank genes within each cluster based on their silhouette
    """

    gene_sil = computeGeneSilhouettes(clusters, data)

    ranked_clusters = {}
    for cluster_id, cluster in clusters.items():
        sil_dict = {gene_id: gene_sil[gene_id] for gene_id in cluster}
        ranked_dict = dict(
            sorted(sil_dict.items(), key=lambda item: item[1], reverse=True)
        )
        ranked_clusters[cluster_id] = ranked_dict

    return ranked_clusters


def writeExcelOfClusterGenes(clusters, out_path, gene_info,
                             gene_pathways, gene_systems):
    """
    Write excel file with gene product and KEGG pathways for genes in each cluster.
    """
    silhouette = type(clusters[list(clusters.keys())[0]]) == dict
    writer = pd.ExcelWriter(out_path, engine='xlsxwriter')

    for cluster_id, cluster in clusters.items():
        cluster_info = {}
        for gene_id in cluster:
            if gene_id in gene_info.keys():
                if silhouette:
                    cluster_info[gene_id] = {
                        'Gene silhouette': cluster[gene_id],
                        'Gene product': gene_info[gene_id],
                        'KEGG system': ', '.join(gene_systems[gene_id]),
                        'KEGG subsystem': ', '.join(gene_pathways[gene_id]),
                    }
                else:
                    cluster_info[gene_id] = {
                        'Gene product': gene_info[gene_id],
                        'KEGG system': ', '.join(gene_systems[gene_id]),
                        'KEGG subsystem': ', '.join(gene_pathways[gene_id]),
                    }

        pd.DataFrame(cluster_info).transpose().to_excel(writer, sheet_name=f'Cluster {cluster_id}')
    writer.save()



import pickle
import random
import numpy as np


def saveToPickleFile(python_object, path_to_file='object.pkl'):
    """
    Save python object to pickle file
    """
    out_file = open(path_to_file, 'wb')
    pickle.dump(python_object, out_file)
    out_file.close()


def readFromPickleFile(path_to_file='object.pkl'):
    """
    Load python object from pickle file.
    Returns python object.
    """
    in_file = open(path_to_file, 'rb')
    python_object = pickle.load(in_file)
    return python_object


def computeSamplePvalue(distribution: list, statistic_value: float) -> float:
    """
    Compute sample p-value given null distribution and statistic value
    """
    N_equal_more_extreme = len(
        np.where(np.array(distribution) >= statistic_value)[0]
        )
    N_total = len(distribution)
    return N_equal_more_extreme / N_total


def copyProxyDictionary(proxydict) -> dict:
    """
    Convert ProxyDict, as returned by multiprocessing.Manager to regular dict.
    """
    return {k: v.copy() for k, v in proxydict.items()}


def randomPartition(elems: list, bin_sizes: list) -> list:
    """
    Randomly partition list elements into
    bins of given sizes
    """
    def shuffleList(elems):
        random.shuffle(elems)
        return elems

    elems = shuffleList(elems)
    partition = []
    start, end = 0, 0
    for bin_size in bin_sizes:
        end += bin_size
        partition.append(elems[start:end])
        start += bin_size
    return partition


def getCounts(array_like: list, sort_by_value=True) -> list:
    """
    Get dict of array item counts. Dict sorted
    by keys unless sort_by_value set to True.
    """
    unique_cond, counts_cond = np.unique(array_like, return_counts=True)
    counts = dict(zip(unique_cond, counts_cond))
    if sort_by_value:
        return dict(sorted(
            counts.items(), key=lambda item: item[1], reverse=True))
    else:
        return counts
